import pandas as pd

import pemi
import pemi.pipes.patterns
import pemi.transforms

class PdForkPipe(pemi.pipes.patterns.ForkPipe):
    def flow(self):
        for target in self.targets.values():
            target.df = self.sources['main'].df.copy()


class PdConcatPipe(pemi.pipes.patterns.ConcatPipe):
    def __init__(self, *, concat_opts={}, **params):
        super().__init__(**params)
        self.concat_opts = concat_opts


    def flow(self):
        source_dfs = [source.df for source in self.sources.values() if source.df is not None]
        if len(source_dfs) == 0:
            self.targets['main'].df = pd.DataFrame()
        else:
            self.targets['main'].df = pd.concat(source_dfs, **self.concat_opts, sort=False)

class PdLookupJoinPipe(pemi.Pipe):
    def __init__(self, main_key, lookup_key,
                 suffixes=('', '_lkp'),
                 on_missing='redirect', #redirect, ignore, warn
                 indicator=None,
                 lookup_prefix='',
                 fillna=None,
                 **kwargs): #pylint: disable=too-many-arguments
        super().__init__(**kwargs)

        self.main_key = main_key
        self.lookup_key = lookup_key
        self.suffixes = suffixes
        self.on_missing = on_missing
        self.indicator = indicator
        self.lookup_prefix = lookup_prefix
        self.fillna = fillna

        self.source(
            pemi.PdDataSubject,
            name='main'
        )

        self.source(
            pemi.PdDataSubject,
            name='lookup'
        )

        self.target(
            pemi.PdDataSubject,
            name='main'
        )

        self.target(
            pemi.PdDataSubject,
            name='errors'
        )


    def _drop_lookup_duplicates(self, _na):
        return self.sources['lookup'].df.drop_duplicates(self.lookup_key)

    def _prefix_lookup_columns(self, lkp_df):
        if self.lookup_prefix == '':
            return lkp_df

        return lkp_df.rename(columns={
            col: self.lookup_prefix + col
            for col in lkp_df.columns if not col in self.lookup_key
        })

    def _drop_missing_lookup_keys(self, lkp_df):
        missing_keys = lkp_df[self.lookup_key].apply(
            lambda v: v.apply(pemi.transforms.isblank).any(), axis=1
        )
        if len(missing_keys) > 0:
            return lkp_df[~missing_keys]
        return lkp_df

    def _perform_lookup_merge(self, lkp_df):
        return pd.merge(
            self.sources['main'].df,
            lkp_df,
            left_on=self.main_key,
            right_on=self.lookup_key,
            how='left',
            suffixes=self.suffixes,
            indicator='__indicator__'
        )

    def _fill_missing(self, merged_df):
        if not self.fillna:
            return merged_df

        merged_df['__indicator__'] = merged_df['__indicator__'].astype('str')
        merged_df.fillna(**self.fillna, inplace=True)
        return merged_df

    def _direct_targets(self, merged_df):
        matches = (merged_df['__indicator__'] == 'both')
        if self.on_missing == 'redirect':
            self.targets['main'].df = merged_df[matches].copy()
            self.targets['errors'].df = merged_df[~matches].copy()
        else:
            self.targets['main'].df = merged_df
            self.targets['errors'].df = pd.DataFrame([], columns=merged_df.columns)

        if self.on_missing == 'warn':
            merged_df[~matches][self.main_key].apply(
                lambda row: pemi.log.warning('No lookup values found for %s', dict(row)),
                axis=1
            )

        return None

    def _remove_indicator(self, _na):
        if self.indicator:
            indicator_bool = self.targets['main'].df['__indicator__'].apply(
                lambda v: True if v == 'both' else False
            ).astype('bool')

            self.targets['main'].df[self.indicator] = indicator_bool

        del self.targets['main'].df['__indicator__']
        del self.targets['errors'].df['__indicator__']

    def flow(self):
        pemi.log.debug('PdLookupJoinPipe - main source columns: %s',
                       self.sources['main'].df.columns)
        pemi.log.debug('PdLookupJoinPipe - main lookup columns: %s',
                       self.sources['lookup'].df.columns)

        arg = None
        for func in [
                self._drop_lookup_duplicates,
                self._prefix_lookup_columns,
                self._drop_missing_lookup_keys,
                self._perform_lookup_merge,
                self._fill_missing,
                self._direct_targets,
                self._remove_indicator
        ]:
            arg = func(arg)


        pemi.log.debug('PdLookupJoinPipe - main target columns: %s',
                       self.targets['main'].df.columns)



class PdFieldValueForkPipe(pemi.Pipe):
    def __init__(self, field, forks, **kwargs):
        super().__init__(**kwargs)

        self.field = field
        self.forks = forks

        self.source(
            pemi.PdDataSubject,
            name='main'
        )

        for fork in self.forks:
            self.target(
                pemi.PdDataSubject,
                name=fork
            )

        self.target(
            pemi.PdDataSubject,
            name='remainder'
        )

    def flow(self):
        grouped = self.sources['main'].df.groupby(self.field)

        for fork in self.forks:
            if fork in grouped.groups:
                self.targets[fork].df = grouped.get_group(fork).copy()
            else:
                self.targets[fork].df = pd.DataFrame(columns=self.sources['main'].df.columns)

        remainder = set(grouped.groups.keys()) - set(self.forks)
        if len(remainder) > 0:
            self.targets['remainder'].df = pd.concat(
                [grouped.get_group(r) for r in remainder]
            ).sort_index()

class PdLambdaPipe(pemi.Pipe):
    '''
    This pipe is used to build quick Pandas transformations where building a full pipe class
    may feel like overkill.  You would use this pipe if you don't need to test it in isolation
    (e.g., it only makes sense in a larger context), or you don't need control over the schemas.

    Args:
      fun (function): A function that accepts a dataframe as argument (source) and returns
        a dataframe (target).

    :Data Sources:
      **main** (*pemi.PdDataSubject*) - The source dataframe that gets pass to ``fun``.

    :Data Targets:
      **main** (*pemi.PdDataSubject*) - The target dataframe that gets populated from
        the return value of ``fun``.

    '''
    def __init__(self, fun):
        super().__init__()

        self.fun = fun

        self.source(
            pemi.PdDataSubject,
            name='main'
        )

        self.target(
            pemi.PdDataSubject,
            name='main'
        )

    def flow(self):
        self.targets['main'].df = self.fun(self.sources['main'].df)



from collections import defaultdict

import dask
import dask.dot

import pemi

class DagValidationError(Exception): pass

class PipeConnection:
    def __init__(self, parent, from_pipe_name, from_subject_name):
        self.parent = parent
        self.from_pipe_name = from_pipe_name
        self.from_subject_name = from_subject_name
        self.group = None
        self.to_pipe_name = None
        self.to_subject_name = None


    def to(self, to_pipe_name, to_subject_name): #pylint: disable=invalid-name
        self.to_pipe_name = to_pipe_name
        self.to_subject_name = to_subject_name
        return self

    @property
    def from_pipe(self):
        return self.parent.pipes[self.from_pipe_name]

    @property
    def to_pipe(self):
        return self.parent.pipes[self.to_pipe_name]

    @property
    def from_subject(self):
        return self.from_pipe.targets[self.from_subject_name]

    @property
    def to_subject(self):
        return self.to_pipe.sources[self.to_subject_name]

    @property
    def from_self(self):
        return self.parent is self.from_pipe

    @property
    def to_self(self):
        return self.parent is self.to_pipe

    def connect(self):
        self.to_subject.connect_from(self.from_subject)

    def group_as(self, name):
        self.group = name
        return self

    def __str__(self):
        return 'PipeConnection: {}.{} -> {}.{}'.format(
            self.from_pipe,
            self.from_subject,
            self.to_pipe,
            self.to_subject
        )

    def __repr__(self):
        return '<{}>'.format(self.__str__())

class DaskPipe: #pylint: disable=too-few-public-methods
    'DaskPipe is just a wrapper around Pipe that allows us to use the Pipes in the context of Dask'

    def __init__(self, pipe, is_parent):
        self.pipe = pipe
        self.is_parent = is_parent

    def __call__(self, *args):
        pemi.log.info('DaskPipe flowing pipe %s', self.pipe)
        if not self.is_parent:
            self.pipe.flow()
        return self

    def __str__(self):
        return self.pipe.name


class PipeConnections:
    def __init__(self, connections=None):
        self.connections = connections or []

    def append(self, conn):
        self.connections.append(conn)

    def group(self, group):
        return self.__class__([c for c in self.connections if c.group == group])

    def flow(self, dask_get=dask.get):
        dask_dag = self.dask_dag()
        return dask_get(dask_dag, list(dask_dag.keys()))


    def graph(self):
        return self.dask_graph()

    def dask_graph(self):
        return dask.dot.dot_graph(self.dask_dag(), rankdir='TB')

    def dask_dag(self):
        self.validate_dag()
        dag = {}
        for conn in self.connections:
            for node, edge in self._node_edge(conn).items():
                if node in dag:
                    dag[node][1].extend(edge[1])
                else:
                    dag[node] = edge

        return dag


    def validate_dag(self):
        targets = defaultdict(list)
        sources = defaultdict(list)
        for conn in self.connections:
            to_str = '{}.sources[{}]'.format(conn.to_pipe_name, conn.to_subject_name)
            from_str = '{}.targets[{}]'.format(conn.from_pipe_name, conn.from_subject_name)
            sources[from_str].append(to_str)
            targets[to_str].append(from_str)

            if not conn.from_pipe_name in conn.parent.pipes:
                msg = "Pipe '{}' not defined in parent pipe '{}'".format(
                    conn.from_pipe_name, conn.parent.name
                )
                raise DagValidationError(msg)
            if not conn.to_pipe_name in conn.parent.pipes:
                msg = "Pipe '{}' not defined in parent pipe '{}'".format(
                    conn.to_pipe_name, conn.parent.name
                )
                raise DagValidationError(msg)
            if not conn.from_subject_name in conn.parent.pipes[conn.from_pipe_name].targets:
                msg = "Pipe '{}' has no target named '{}'".format(
                    conn.from_pipe_name, conn.from_subject_name
                )
                raise DagValidationError(msg)
            if not conn.to_subject_name in conn.parent.pipes[conn.to_pipe_name].sources:
                msg = "Pipe '{}' has no source named '{}'".format(
                    conn.from_pipe_name, conn.from_subject_name
                )
                raise DagValidationError(msg)

        target_dupes = {k:v for k, v in targets.items() if len(v) > 1}
        if len(target_dupes) > 0:
            msg = 'Multiple connections to the same target.  ' \
                + 'Use a concatenator pipe instead.  Details: {}'.format(target_dupes)
            raise DagValidationError(msg)

        source_dupes = {k:v for k, v in sources.items() if len(v) > 1}
        if len(source_dupes) > 0:
            msg = 'Multiple connections from the same source.  ' \
                + 'Use a fork pipe instead.  Details: {}'.format(source_dupes)
            raise DagValidationError(msg)

    @staticmethod
    def _connect_to(source, connection):
        def __connect_to(target):
            pemi.log.debug('connecting %s to %s', target, source)
            connection.connect()
            return target
        __connect_to.__name__ = 'connect_to'
        return __connect_to

    @staticmethod
    def _get_target(name):
        def __get_target(daskpipe):
            pemi.log.debug('Getting target %s from pipe %s', name, daskpipe.pipe)
            return daskpipe.pipe.targets[name]
        __get_target.__name__ = '[{}]'.format(name)
        return __get_target

    def _node_edge(self, conn):
        return {
            '{}.targets'.format(conn.from_pipe_name):
                (DaskPipe(conn.from_pipe, conn.from_self), []),
            '{}.targets[{}]'.format(conn.from_pipe_name, conn.from_subject_name):
                (self._get_target(conn.from_subject_name), '{}.targets'.format(
                    conn.from_pipe_name)),
            '{}.sources[{}]'.format(conn.to_pipe_name, conn.to_subject_name):
                (self._connect_to(conn.to_pipe.sources[conn.to_subject_name], conn),
                 '{}.targets[{}]'.format(conn.from_pipe_name, conn.from_subject_name)),
            '{}.targets'.format(conn.to_pipe_name):
                (DaskPipe(conn.to_pipe, conn.to_self),
                 ['{}.sources[{}]'.format(conn.to_pipe_name, conn.to_subject_name)])
        }



import math
import numpy as np


def isblank(value):
    try:
        return math.isnan(value)
    except TypeError:
        blank = [np.nan, np.datetime64("NaT"), [], {}, None, '']
        return value in blank


def concatenate(delimiter=''):
    def _concatenate(row):
        return delimiter.join(row)

    return _concatenate


def nvl(default=''):
    def _nvl(row):
        return next((v for v in row if not isblank(v)), default)

    return _nvl



import os
import re

import pandas as pd

import pemi
from pemi.pipes.patterns import TargetPipe

def default_column_normalizer(name):
    name = str(name)
    name = re.sub(r'\s+', '_', name).lower()
    name = re.sub(r'[^a-z0-9_]', '_', name)
    return name


class StrConverter(dict):
    def __contains__(self, key):
        return True

    def __getitem__(self, key):
        return str

class LocalCsvFileSourcePipe(pemi.Pipe):
    def __init__(self, *, paths, schema=None, csv_opts=None,
                 filename_field=None, filename_full_path=False, normalize_columns=True, **params):
        super().__init__(**params)

        self.paths = paths
        self.schema = schema
        self.filename_field = filename_field
        self.filename_full_path = filename_full_path
        self.csv_opts = self._build_csv_opts(csv_opts or {})

        if callable(normalize_columns):
            self.column_normalizer = normalize_columns
        elif normalize_columns:
            self.column_normalizer = default_column_normalizer
        else:
            self.column_normalizer = lambda col: col

        self.target(
            pemi.PdDataSubject,
            name='main',
            schema=self.schema
        )

        self.target(
            pemi.PdDataSubject,
            name='errors',
            schema=self.schema
        )

    def extract(self):
        return self.paths

    def parse(self, data):
        pemi.log.debug('Parsing files at %s', data)

        filepaths = data
        mapped_dfs = []
        error_dfs = []
        for filepath in filepaths:
            parsed_dfs = self._parse_one(filepath)
            mapped_dfs.append(parsed_dfs.mapped)
            error_dfs.append(parsed_dfs.errors)

        if len(filepaths) > 0:
            self.targets['main'].df = pd.concat(mapped_dfs)
            self.targets['errors'].df = pd.concat(error_dfs)
        elif self.schema:
            self.targets['main'].df = pd.DataFrame(columns=list(self.schema.keys()))
            self.targets['errors'].df = pd.DataFrame(columns=list(self.schema.keys()))
        else:
            self.targets['main'].df = pd.DataFrame()
            self.targets['errors'].df = pd.DataFrame()

        pemi.log.debug('Parsed %i records', len(self.targets['main'].df))
        return self.targets['main'].df

    def _build_csv_opts(self, user_csv_opts):
        if self.schema:
            file_fieldnames = [k for k in self.schema.keys() if k != self.filename_field]
            usecols = lambda col: self.column_normalizer(col) in file_fieldnames
        else:
            usecols = None

        mandatory_opts = {
            'converters': StrConverter(),
            'usecols': usecols
        }

        default_opts = {
            'engine':          'c',
            'error_bad_lines': True
        }

        return {**default_opts, **user_csv_opts, **mandatory_opts}

    def _parse_one(self, filepath):
        pemi.log.debug('Parsing file at %s', filepath)

        raw_df = pd.read_csv(filepath, **self.csv_opts)
        pemi.log.debug('Found %i raw records', len(raw_df))

        raw_df.columns = [self.column_normalizer(col) for col in raw_df.columns]

        if self.filename_field:
            if self.filename_full_path:
                raw_df[self.filename_field] = filepath
            else:
                raw_df[self.filename_field] = os.path.basename(filepath)

        if self.schema:
            return raw_df.mapping(
                [(name, name, field.coerce) for name, field in self.schema.items()],
                on_error='redirect'
            )
        return raw_df.mapping([], inplace=True)

    def flow(self):
        self.parse(self.extract())


class LocalCsvFileTargetPipe(TargetPipe):
    def __init__(self, *, path, csv_opts=None, **params):
        super().__init__(**params)

        self.path = path
        self.csv_opts = self._build_csv_opts(csv_opts or {})

    def encode(self):
        return self.sources['main'].df

    def load(self, encoded_data):
        df = encoded_data
        df.to_csv(self.path, **self.csv_opts)
        return self.path

    @staticmethod
    def _build_csv_opts(user_csv_opts):
        mandatory_opts = {}

        default_opts = {
            'index': False
        }

        return {**default_opts, **user_csv_opts, **mandatory_opts}



'''
Package containing core Pemi functionality
'''
import logging
#pylint: disable=wrong-import-position
log = logging.getLogger('pemi') #pylint: disable=invalid-name
log.setLevel(logging.WARN)

work_dir = '.' #pylint: disable=invalid-name
#pylint: enable=wrong-import-position

import pandas
import pandas_mapper

from pemi.version import __version__ as version


import pemi.schema
from pemi.schema import Schema

import pemi.pipe
from pemi.pipe import Pipe

import pemi.fields
from pemi.fields import *

import pemi.data_subject
from pemi.data_subject import *

import pemi.transforms



from collections import OrderedDict
from itertools import tee

import os
import sys

import pytest
import pandas as pd

import pemi
import pemi.data

# Imported here for backwards compatiblity
from pemi.pipe import mock_pipe #pylint: disable=unused-import

pd.set_option('display.expand_frame_repr', False)

def assert_frame_equal(actual, expected, **kwargs):
    try:
        pd.util.testing.assert_frame_equal(actual, expected, **kwargs)
    except AssertionError as err:
        msg = str(err)
        msg += '\nActual:\n{}'.format(actual)
        msg += '\nExpected:\n{}'.format(expected)
        raise AssertionError(msg)

def assert_series_equal(actual, expected, **kwargs):
    actual.reset_index(drop=True, inplace=True)
    expected.reset_index(drop=True, inplace=True)
    try:
        pd.util.testing.assert_series_equal(actual, expected, **kwargs)
    except AssertionError as err:
        msg = str(err)
        msg += '\nActual:\n{}'.format(actual)
        msg += '\nExpected:\n{}'.format(expected)
        raise AssertionError(msg)

class when: #pylint: disable=invalid-name
    #pylint: enable=invalid-name

    @staticmethod
    def source_has_keys(source, case_keys):
        def gen_values(case):
            while True:
                values = {}
                cache_id = os.urandom(32)
                for field in case_keys.subject_keys[source.name]:
                    values[field] = case_keys.cache(source.name, field, case=case)[cache_id]
                yield values

        def _when(case):
            values = gen_values(case)
            keys_df = pd.DataFrame([next(values) for i in range(len(source[case].data))])
            for field in case_keys.subject_keys[source.name]:
                source[case].data[field] = keys_df[field]

        return _when

    @staticmethod
    def source_field_has_value(source, field, value):
        def _when(case):
            nrecords = len(source[case].data)
            if hasattr(value, '__next__'):
                source[case].data[field] = pd.Series([next(value) for i in range(nrecords)])
            else:
                source[case].data[field] = pd.Series([value]*nrecords)

        return _when

    @staticmethod
    def source_fields_have_values(source, mapping):
        def _when(case):
            for field, value in mapping.items():
                nrecords = len(source[case].data)
                if hasattr(value, '__next__'):
                    source[case].data[field] = pd.Series([next(value) for i in range(nrecords)])
                else:
                    source[case].data[field] = pd.Series([value]*nrecords)

        return _when

    @staticmethod
    def example_for_source(source, table):
        def _when(case):
            source[case].data = table.df

        return _when

    @staticmethod
    def source_conforms_to_schema(source):
        def _when(case):
            data = pemi.data.Table(
                nrows=3,
                schema=source.subject.schema
            ).df

            source[case].data = data
        return _when


class then: #pylint: disable=invalid-name
    #pylint: enable=invalid-name

    @staticmethod
    def target_field_has_value(target, field, value):
        def _then(case):
            target_data = pd.Series(list(target[case].data[field]))
            expected_data = pd.Series([value] * len(target_data), index=target_data.index)

            assert_series_equal(target_data, expected_data,
                                check_names=False,
                                check_dtype=False,
                                check_datetimelike_compat=True)

        return _then

    @staticmethod
    def target_fields_have_values(target, mapping):
        def _then(case):
            actual = target[case].data[list(mapping.keys())]
            expected = pd.DataFrame(index=actual.index)
            for k, v in mapping.items():
                expected[k] = pd.Series([v] * len(actual), index=actual.index)

            assert_frame_equal(actual, expected, check_names=False, check_dtype=False)
        return _then

    @staticmethod
    def target_matches_example(target, expected_table, by=None):
        subject_fields = expected_table.defined_fields

        def _then(case):
            expected = expected_table.df[subject_fields]
            actual = target[case].data[subject_fields]

            if by:
                expected = expected.sort_values(by).reset_index(drop=True)
                actual = actual.sort_values(by).reset_index(drop=True)
            else:
                expected = expected.reset_index(drop=True)
                actual = actual.reset_index(drop=True)

            assert_frame_equal(actual, expected, check_names=False, check_dtype=False)

        return _then

    @staticmethod
    def field_is_copied(source, source_field, target, target_field, by=None, #pylint: disable=too-many-arguments
                        source_by=None, target_by=None):
        source_by = source_by or by
        target_by = target_by or by or source_by

        def _then(case):
            if source_by:
                expected = source[case].data.sort_values(source_by)\
                                            .reset_index(drop=True)[[source_field]]
                actual = target[case].data.sort_values(target_by)\
                                          .reset_index(drop=True)[[target_field]]
            else:
                expected = source[case].data[[source_field]]
                actual = target[case].data[[target_field]]

            try:
                assert_series_equal(expected[source_field], actual[target_field], check_names=False)
            except AssertionError as err:
                raise AssertionError(
                    'Source field {} not copied to target field {}: {}'.format(
                        source_field, target_field, err
                    )
                )

        return _then

    @staticmethod
    def fields_are_copied(source, target, mapping, by=None, source_by=None, target_by=None): #pylint: disable=too-many-arguments
        source_fields = list(set([m[0] for m in mapping]))
        target_fields = list(set([m[1] for m in mapping]))

        source_by = source_by or by
        target_by = target_by or by or source_by

        def _then(case):
            if source_by:
                expected = source[case].data.sort_values(source_by)\
                                            .reset_index(drop=True)[source_fields]
                actual = target[case].data.sort_values(target_by)\
                                          .reset_index(drop=True)[target_fields]
            else:
                expected = source[case].data[source_fields]
                actual = target[case].data[target_fields]

            for source_field, target_field in mapping:
                try:
                    assert_series_equal(expected[source_field], actual[target_field],
                                        check_names=False, check_dtype=False)
                except AssertionError as err:
                    raise AssertionError(
                        'Source field {} not copied to target field {}: {}'.format(
                            source_field, target_field, err
                        )
                    )

        return _then


    @staticmethod
    def target_does_not_have_fields(target, fields):
        def _then(case):
            unexpected_fields = set(fields) & set(target[case].data.columns)
            if len(unexpected_fields) > 0:
                raise AssertionError(
                    "The fields '{}' were not expected to be found in the target".format(
                        unexpected_fields
                    )
                )

        return _then

    @staticmethod
    def target_has_fields(target, fields, only=False):
        def _then(case):
            missing_fields = set(fields) - set(target[case].data.columns)
            extra_fields = set(target[case].data.columns) - set(fields)

            if len(missing_fields) > 0:
                raise AssertionError(
                    "The fields '{}' were expected to be found in the target".format(
                        missing_fields
                    )
                )

            if len(extra_fields) > 0 and only:
                raise AssertionError(
                    "The fields '{}' were not expected to be found on the target".format(
                        extra_fields
                    )
                )

        return _then

    @staticmethod
    def target_is_empty(target):
        def _then(case):
            nrecords = len(target[case].data)
            if nrecords != 0:
                raise AssertionError(
                    'Expecting target to be empty, found {} records'.format(
                        nrecords
                    )
                )

        return _then

    @staticmethod
    def target_has_n_records(target, expected_n):
        def _then(case):
            nrecords = len(target[case].data)
            if nrecords != expected_n:
                raise AssertionError(
                    'Excpecting target to have {} records, found {} records'.format(
                        expected_n, nrecords
                    )
                )

        return _then




class SubscriptableLambda: #pylint: disable=too-few-public-methods
    def __init__(self, func):
        self.func = func

    def __getitem__(self, cache=None):
        return self.func(cache)


class CaseKeyTracker:
    def __init__(self, case_key_gen):
        self.case = None

        self.cached = {}
        self.subject_case_keys = {}

        self.subject_keys = {name: list(keys.keys()) for name, keys in next(case_key_gen).items()}
        self._build_synced_generators(case_key_gen)

    def next_case(self, case):
        self.cached = {}
        self.case = case

    def _track_key(self, subject_case_keys, case=None):
        case = case if case else self.case
        for subject_name, key in subject_case_keys.items():
            if subject_name not in self.subject_case_keys:
                self.subject_case_keys[subject_name] = []
            self.subject_case_keys[subject_name].append({'case': case, 'key': key})

    def _build_synced_generators(self, case_key_gen):
        subjects = list(self.subject_keys.keys())

        self.synced_generators = {}
        for idx, tgen in enumerate(tee(case_key_gen, len(subjects) + 1)):
            if idx < len(subjects):
                self.synced_generators[subjects[idx]] = tgen
            else:
                self.synced_generators['__internal__'] = tgen

    def get_key(self, subject, field, case=None, cache=None):
        cache = os.urandom(32) if cache is None else cache
        if cache not in self.cached:
            self.cached[cache] = next(self.synced_generators['__internal__'])
            self._track_key(self.cached[cache], case=case)

        try:
            return self.cached[cache][subject][field]
        except KeyError:
            msg = "No case key found for field '{}' and subject '{}'".format(field, subject)
            raise KeyError(msg)

    def cache(self, subject, field, case=None):
        return SubscriptableLambda(lambda name: self.get_key(subject, field, case=case, cache=name))



class CaseData: #pylint: disable=attribute-defined-outside-init,too-few-public-methods
    def __init__(self, case, test_subject):
        self.case = case
        self.test_subject = test_subject

    @property
    def data(self):
        if not hasattr(self, '_data'):
            cols = self.test_subject.subject.schema.keys()
            self._data = pd.DataFrame([[None] * len(cols)], columns=cols)

        return self._data

    @data.setter
    def data(self, value):
        self._data = value

class TestSubject: #pylint: disable=too-few-public-methods
    def __init__(self, subject, name):
        self.subject = subject
        self.name = name
        self.data = {}

    def __getitem__(self, case):
        case_id = id(case)
        if case_id not in self.data:
            self.data[case_id] = CaseData(case, self)
        return self.data[case_id]

class Case:
    def __init__(self, name, scenario):
        self.name = name
        self.scenario = scenario
        self.whens = []
        self.thens = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def when(self, *funcs):
        self.whens.extend(funcs)
        return self

    def then(self, *funcs):
        self.thens.extend(funcs)
        return self

    def setup(self):
        for i_when in self.whens:
            i_when(self)

    def assert_case(self):
        self.scenario.run()

        try:
            for i_then in self.thens:
                i_then(self)
        except AssertionError:
            msg = '\nAssertion Error for {}'.format(self)
            for name, source in self.scenario.sources.items():
                msg += '\nSource {}:\n{}'.format(name, source.subject.to_pd())
            for name, target in self.scenario.targets.items():
                msg += '\nTarget {}:\n{}'.format(name, target.subject.to_pd())
            raise AssertionError(msg)


    def __str__(self):
        return "<Case '{}' ({})>".format(self.name, id(self))

class Scenario: #pylint: disable=too-many-instance-attributes
    def __init__(self, name, *selector, usefixtures=[]):
        self.name = name
        self.selector = selector
        self.usefixtures = usefixtures

        self.runner = None
        self.case_keys = None

        self.cases = OrderedDict()
        self.has_run = False

        self.sources = None
        self.targets = None


    def _register_test(self, module_name):
        @pytest.mark.usefixtures(*self.usefixtures)
        @pytest.mark.scenario(self, *self.selector)
        def test_scenario(case):
            case.assert_case()
        setattr(sys.modules[module_name], 'testScenario:{}'.format(self.name), test_scenario)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        import inspect
        current_frame = inspect.currentframe()
        calling_module = inspect.getouterframes(current_frame)[1].frame.f_locals['__name__']
        self._register_test(calling_module)

    def setup(self, runner, case_keys=None, sources=None, targets=None):
        sources = sources or {}
        targets = targets or {}

        self.runner = runner
        self.case_keys = CaseKeyTracker(case_keys)

        self.sources = {name: TestSubject(subject, name) for name, subject in sources.items()}
        self.targets = {name: TestSubject(subject, name) for name, subject in targets.items()}


    def case(self, name):
        case = Case(name, self)
        self.case_keys.next_case(case)
        self.cases[name] = case
        return case

    def setup_cases(self):
        for case in self.cases.values():
            case.setup()

        self.load_test_data()

    def load_test_data(self):
        for _, source in self.sources.items():
            if len(source.data.values()) > 0:
                all_case_data = pd.concat(
                    [cd.data for cd in source.data.values()],
                    ignore_index=True,
                    sort=False
                )
                source.subject.from_pd(all_case_data)


    def collect_results(self):
        def assign_case(target_name):
            try:
                subject_case_keys = self.case_keys.subject_case_keys[target_name]
            except KeyError:
                raise KeyError('''No case keys assigned for target '{}'.
                     This usually means you forgot to build source or target case keys.
                     '''.format(target_name))

            key_fields = list(subject_case_keys[0]['key'].keys())

            def _assign_case(row):
                row_key = set(dict(row[key_fields]).items())
                for key_case in subject_case_keys:
                    key = set(key_case['key'].items())
                    if len(row_key & key) > 0:
                        return key_case['case']
                raise KeyError("No case found for {} for target '{}'".format(row_key, target_name))
            return _assign_case

        for target_name, target in self.targets.items():
            all_target_data = target.subject.to_pd()
            for case in self.cases.values():
                target[case].data = pd.DataFrame(columns=all_target_data.columns)

            if len(all_target_data) > 0:
                all_target_data['__pemi_case__'] = all_target_data.apply(
                    assign_case(target_name), axis=1
                )
                for case, df in all_target_data.groupby(['__pemi_case__'], sort=False):
                    del df['__pemi_case__']
                    target[case].data = df

    def assert_cases(self):
        for case in self.cases.values():
            case.assert_case()

    def run(self):
        if self.has_run:
            return

        self.setup_cases()
        self.runner()
        self.collect_results()

        self.has_run = True
